// 
// Copyright (c) 2010, Benjamin Kaufmann
// 
// This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
// 
// Clasp is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Clasp is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Clasp; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
#include "claspre.h"
#if defined(WITH_CLASPRE_SUPPORT)
#include <program_opts/value.h>
#include "clasp_output.h"
#include "clasp_options.h"
#include <stdio.h>
#include <sstream>
#include <numeric>
namespace Clasp { namespace Claspre {
bool Options::mapFormat(const std::string& s, int& format) {
	std::string temp = ProgramOptions::toLower(s);
	bool b;
	if (ProgramOptions::parseValue(temp, b, 1)) {
		format = b ? features_verbose : features_no;
		return true;
	}
	format = -1;
	if      (temp == "v")  { format = features_verbose; }
	else if (temp == "c1") { format = features_compact_1; }
	else if (temp == "c2") { format = features_compact_2; }
	else if (temp == "c3") { format = features_compact_3; }
	return format != -1;
}
void Options::initOptions(ProgramOptions::OptionGroup& root) {
	using namespace ProgramOptions;
	OptionGroup pre("Claspre Options");
	pre.addOptions()
		("list-features"   , bool_switch(&listFeatures),"Print feature names (in --features=C1 format/order)")
		("features"   , storeTo(features)->parser(&Options::mapFormat),
			"Print features in selected format\n"
			"      Default: no\n"
			"      Valid: V, C1, C2, no\n"
			"        V : Print features in human-readable format\n"
			"        C1: Print in compact format; iterated features in separate lines\n"
			"        C2: Print in compact format; iterated features all in one line\n"
			"        C3: Print in C2 format but only if search state is unknown on exit\n")
	;
	root.addOptions(pre, false);
}

class ClaspreOutput : public AspOutput, public Clasp::Enumerator::ProgressReport {
public:
	ClaspreOutput(Options::FeatureFormat f, int& verb, bool asp09);
	void initSolve(const Solver& s, ProgramBuilder* api, Enumerator* e);
	void printSolution(const Solver& s, const Enumerator& en, bool complete);
	void reportRestart(const Solver& s, uint64 maxC, uint32 maxL);
private:
	std::string iterationBuffer_;
	Options::FeatureFormat format_;
	uint32 starts_;
	int&   verbose_;
};

ClaspreOutput::ClaspreOutput(Options::FeatureFormat f, int& v, bool asp09) : AspOutput(asp09), format_(f), starts_(0), verbose_(v) {
}

void ClaspreOutput::initSolve(const Solver& s, ProgramBuilder* api, Enumerator* e) {
	AspOutput::initSolve(s, api, e);
	starts_          = 0;
	iterationBuffer_ = "";
	e->enableProgressReport(this);
}

void ClaspreOutput::printSolution(const Solver& s, const Enumerator& en, bool complete) {
	bool hasSolution = complete || s.stats.solve.models > 0;
	if (format_ == Options::features_verbose) {
		AspOutput::printSolution(s, en, complete);
	}
	else {
		if (starts_ == 0) {
			reportRestart(s, 0, 0);
		}
		if (!iterationBuffer_.empty()) {
			printf("\n%s\n", iterationBuffer_.c_str());
		}
		const PreproStats& ps = programStats();
		uint32 auxAtoms = ps.trStats ? ps.trStats->auxAtoms : 0;
		uint32 orgAtoms = ps.atoms-auxAtoms;
		uint32 extRules = ps.rules[2] + ps.rules[5];
		if (ps.trStats) {
			extRules     -= (ps.trStats->rules[2]+ps.trStats->rules[5]);
		}
		if (!extRules) extRules = 1;
		printf("%s,%u,%u,%u,%u,%u,%u,%u,%u,%.3f", (hasSolution ? "Yes":"No"), ps.atoms, orgAtoms, auxAtoms, ps.rules[0], ps.rules[1]
			, ps.rules[2]-(ps.trStats?ps.trStats->rules[2]: 0)
			, ps.rules[3]-(ps.trStats?ps.trStats->rules[3]: 0)
			, ps.rules[5]-(ps.trStats?ps.trStats->rules[5]: 0)
			, ps.rules[1] / static_cast<double>(extRules));
		uint32 tight = ps.sccs == 0 || ps.sccs == PrgNode::noScc;
		uint32 sccs  = ps.sccs != PrgNode::noScc ? ps.sccs : 0;
		printf(",%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%.3f,%.3f,%.3f", ps.bodies, ps.sumEqs(), ps.numEqs(Var_t::atom_var), ps.numEqs(Var_t::body_var), ps.numEqs(Var_t::atom_body_var)
			, tight, sccs, ps.ufsNodes, s.stats.problem.vars, s.stats.problem.eliminated
			, s.stats.problem.constraints[0]
			, percent(s.stats.problem.constraints[1], s.stats.problem.constraints[0])
			, percent(s.stats.problem.constraints[2], s.stats.problem.constraints[0])
			, percent(s.stats.problem.constraints[0] - (s.stats.problem.constraints[1]+s.stats.problem.constraints[2]), s.stats.problem.constraints[0]));
		std::stringstream temp;
		const SolveStats& st  = s.stats.solve;
		const JumpStats& js   = s.stats.solve.jumps   ? *s.stats.solve.jumps   : JumpStats();
		uint64 learntSum      = std::accumulate(st.learnts, st.learnts+Constraint_t::max_value, uint64(0));
		temp.precision(3);
		temp << std::fixed << "," << st.models << "," << st.choices << "," << st.conflicts << "," << st.restarts << "," << st.deleted << "," 
			   << (st.conflicts-js.jumps) << "," << js.jumps << "," << js.bJumps << ","
				 << js.jumpSum << "," << (js.jumpSum - js.boundSum) << "," << percent(js.jumpSum - js.boundSum, js.jumpSum) << ","
				 << js.maxJump << "," << js.maxJumpEx << "," << js.maxBound << "," << js.avgJumpLen() << "," << js.avgJumpLenEx() << ","
				 << average(js.boundSum,js.bJumps) << "," << average(st.modLits,st.models) << ","
		     << learntSum << "," << percent(st.binary, learntSum) << "," << percent(st.ternary, learntSum) << "," 
				 << percent(learntSum-st.binary-st.ternary, learntSum) << ","
				 << st.learnts[0] << "," << average(st.lits[0], st.learnts[0]) << ","
				 << st.learnts[1] << "," << average(st.lits[1], st.learnts[1]);
		printf("%s\n\n", temp.str().c_str());
		if (format_ == Options::features_compact_3 && hasSolution) {
			AspOutput::printSolution(s, en, complete);
		}
		else {
			verbose_ = 0;
		}
	}
}

void ClaspreOutput::reportRestart(const Solver& s, uint64 maxC, uint32 maxL) {
	std::stringstream temp;
	if (format_ == Options::features_verbose) {
		printf("\n[Iteration %u]\n", starts_);
		printf("%-22s: %u\n", "maxLearnt", maxL);
		printf("%-22s: %"PRIu64"\n", "maxConflicts", maxC);
		printf("%-22s: %u\n", "Constraints", s.numConstraints());
		printf("%-22s: %u\n", "LearntConstraints", s.numLearntConstraints());
		printf("%-22s: %u\n", "FreeVars", s.numFreeVars());
		printf("%-22s: %.3f\n", "Vars/FreeVars", ((double)s.numVars())/std::max(s.numFreeVars(),uint32(1)));
		printf("%-22s: %.3f\n", "FreeVars/Constraints", ((double)s.numFreeVars())/s.numConstraints());
		printf("%-22s: %.3f\n", "Vars/Constraints", ((double)s.numVars())/s.numConstraints());
		printf("%-22s: %.3f\n", "maxLearnt/Constraints", ((double)maxL)/s.numConstraints());
		if (s.stats.solve.jumps) {
			printf("%-22s: %.3f\n", "Average Jump Length", s.stats.solve.jumps->avgJumpLen());
		}
		printf("=================================\n\n");
	}
	else {
		uint32 numC = std::max(s.numConstraints(), uint32(1));
		temp.precision(3);
		temp << std::fixed << maxL << "," << maxC << "," << s.numConstraints() << "," << s.numLearntConstraints() << "," << s.numFreeVars() << ","
			   << (static_cast<double>(s.numVars())/std::max(s.numFreeVars(), uint32(1))) << ","
				 << (static_cast<double>(s.numFreeVars())/numC) << ","
				 << (static_cast<double>(s.numVars())/numC) << ","
				 << (static_cast<double>(maxL)/numC);
		if (format_ == Options::features_compact_1) {
			printf("%s\n", temp.str().c_str());
		}
		else {
			if (!iterationBuffer_.empty()) iterationBuffer_ += ",";
			iterationBuffer_ += temp.str();
			verbose_ = format_ == Options::features_compact_3;
		}
	}
	++starts_;
}

OutputFormat* createOutput(Solver& s, Options::FeatureFormat f, ClaspOptions& opts, int& verbose) {
	assert(f != 0);
	s.stats.solve.enableJumpStats();
	if (f == Options::features_verbose) {
		opts.basic.stats = 2;
	}
	else {
		opts.basic.stats = 0;
		verbose          = 0;
		if (f != Options::features_compact_3) {
			opts.basic.quiet = true;
		}
	}
	return new ClaspreOutput(static_cast<Options::FeatureFormat>(f), verbose, opts.basic.asp09);
}


void printFeatures() {
	printf("maxLearnt,maxConflicts,Constraints,LearntConstraints,FreeVars,Vars/FreeVars,FreeVars/Constraints,Vars/Constraints,maxLearnt/Constraints\n");
	printf(
		"Completed,,Atoms,_Original,_Auxiliary,Rules,_Basic,_Constraint,_Choice,_Weight,NormalRules/ExtRules"
		",Bodies,Equivalences,_Atom=Atom,_Body=Body,_Other,Tight,_SCCs,_Nodes,Variables,_Eliminated,Constraints,_Binary,_Ternary,_Other"
		",Models,Choices,Conflicts,Restarts,Constraints deleted,Backtracks,Backjumps,_Bounded,Skippable Levels,Levels skipped,_%%,Max Jump Length"
		",_Executed,Max Bound Length,Average Jump Length,_Executed,Average Bound Length,Average Model Length,Lemmas,_Binary,_Ternary,_Other,Conflicts,_Average Length,Loops,_Average Length\n"
	);
}
}}
#endif // WITH_CLASPRE_SUPPORT
